/* pro_clk.c: real time clock

   Copyright (c) 1997-2003, Tarik Isani (xhomer@isani.org)

   This file is part of Xhomer.

   Xhomer is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 
   as published by the Free Software Foundation.

   Xhomer is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Xhomer; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/* TBD:
		-some bits should not be cleared by reset?
		-BCD mode not implemented
		-12 hour mode not implemented
		-note: CATCHUP int is commented out
		-alarm int could generate extra int, or none at all
*/

/* When enabled, generates interrupts at the following rates
   in "real" PRO time (or actual realtime if RTC is defined):

   RS	Frequency (Hz)
   --   --------------
    0   none
    1    256
    2    128
    3   8192
    4   4096
    5   2048
    6   1024
    7    512
    8    256
    9    128
   10     64
   11     32
   12     16
   13      8
   14      4
   15      2
*/

#ifdef PRO
#include "pdp11_defs.h"
#include "sim_defs.h" /* For sim_gtime() */

#include <time.h>
#include <sys/time.h>
#include <unistd.h>

int		pro_force_year = -1;	/* when > -1, will override host's RTC year value */

LOCAL int	pro_clk_csr0;
LOCAL int	pro_clk_csr1;
LOCAL int	pro_clk_csr2;
LOCAL int	pro_clk_csr3;

LOCAL int	pro_clk_al_sec;		/* alarm registers */
LOCAL int	pro_clk_al_min;
LOCAL int	pro_clk_al_hour;

LOCAL int	pro_clk_rs;		/* freq. # for previous periodic int cycle */
LOCAL int	pro_clk_realtime;	/* indicates whether in realtime mode */
LOCAL int	pro_clk_numint;		/* number of outstanding interrupts */
LOCAL int	pro_clk_freq;		/* periodic int frequency in Hz */
LOCAL int	pro_clk_int_cnt;	/* interrupt counter used for update cycles */

LOCAL struct tm	*pro_time;

LOCAL int	pro_time_sec;
LOCAL int	pro_time_min;
LOCAL int	pro_time_hour;

/* XXX should be int64 */

LOCAL long long 	pro_time_start, pro_time_current;

/* Periodic interrupt frequencies (in Hz) */

LOCAL int	pro_clk_intf[16] = {8192, 256, 128, 8192, 4096, 2048, 1024, 512,
	                    256, 128, 64, 32, 16, 8, 4, 2};


#ifdef RTC
/* Return time in microseconds */

/* XXX should be int64 */

LOCAL long long pro_time_get()
{
	struct timeval	tv;

	long long 	tus;

	gettimeofday(&tv, NULL);
	tus = tv.tv_sec * 1e6 + tv.tv_usec;

	return tus;
}
#endif


/* Update pro_time variable */

LOCAL void pro_time_update()
{
	time_t	t;

	t = time(NULL);
	pro_time = localtime(&t);

	pro_time_sec = pro_time->tm_sec;
	pro_time_min = pro_time->tm_min;
	pro_time_hour = pro_time->tm_hour;

	/* Ignore leap seconds */

	if (pro_time_sec > 59)
	  pro_time_sec = 59;
}


/* Event scheduler */

LOCAL void pro_clk_sched ()
{
int interval, rs;

	/* Recalculate frequency and start time if RS has changed */

	rs = pro_clk_csr0 & PRO_CLK_RS;

	if (rs != pro_clk_rs)
	{
	  pro_clk_rs = rs;

	  /* Update pro_clk_freq */

	  pro_clk_freq = pro_clk_intf[pro_clk_rs];
#ifdef RTC
	  /* Initialize time register */

	  pro_time_start = pro_time_get() * pro_clk_freq;
#endif
	  /* Clear number of outstanding interrupts */

	  pro_clk_numint = 0;

	  /* Clear interrupt counter */

	  pro_clk_int_cnt = 0;
	}

	/* Determine delay interval until next interrupt */

	interval = PRO_EQ_CLK_IPS / pro_clk_freq;

#ifdef RTC
	/* Use shorter interval if more than 2 interrupts are outstanding
	   This will allow the RTC to catch up */

/* XXX This seems to cause a problem

	if (pro_clk_numint > 2)
	  interval = PRO_CLK_RTC_CATCHUP;
*/
#endif
	/* Schedule event */

	pro_eq_sched(PRO_EVENT_CLK, interval);
}

/* Update-ended event scheduler */

LOCAL void pro_clk_uf_sched ()
{
	pro_eq_sched(PRO_EVENT_CLK_UF, PRO_EQ_CLK_UF);
}

/* Update-ended event handler */

void pro_clk_uf_eq ()
{
	/* Update time and date */
 
	pro_time_update();

	/* Clear update-in-progress bit */

	pro_clk_csr0 = pro_clk_csr0 & ~PRO_CLK_UIP;

	/* Set UF bit */

	pro_clk_csr2 = pro_clk_csr2 | PRO_CLK_UF;

	/* Trigger update-ended interrupt, if enabled */

	if ((pro_clk_csr1 & PRO_CLK_UIE) != 0)
	{
	  pro_clk_csr2 = pro_clk_csr2 | PRO_CLK_IRQF;
	  pro_int_set(PRO_INT_CLK);
	}

	/* Compare against alarm registers and trigger
	   alarm interrupt if appropriate */

	if (((pro_time_sec == pro_clk_al_sec)
	   || ((pro_clk_al_sec & PRO_CLK_AL_X) == PRO_CLK_AL_X))
	   && ((pro_time_min == pro_clk_al_min)
	   || ((pro_clk_al_min & PRO_CLK_AL_X) == PRO_CLK_AL_X))
	   && ((pro_time_hour == pro_clk_al_hour)
	   || ((pro_clk_al_hour & PRO_CLK_AL_X) == PRO_CLK_AL_X)))
	{
	  /* Set AF bit */

	  pro_clk_csr2 = pro_clk_csr2 | PRO_CLK_AF;

	  /* Trigger alarm interrupt, if enabled */

	  if ((pro_clk_csr1 & PRO_CLK_AIE) != 0)
	  {
	    pro_clk_csr2 = pro_clk_csr2 | PRO_CLK_IRQF;
	    pro_int_set(PRO_INT_CLK);
	  }
	}
}

/* Periodic interrupt event queue handler */

void pro_clk_eq ()
{
int	set_int;
#ifdef LIMIT
unsigned long	num_us;		/* number uS to sleep for realtime operation */
#endif

	/* Set interrupt and schedule next interrupt, if enabled */

#ifdef RTC
	/* Check if enough instructions have been simulated
	   to switch to realtime mode */

	if (pro_clk_realtime == 0)
	{
	  set_int = 1;

	  if (sim_gtime() > PRO_CLK_RTC_SWITCH)
	    pro_clk_realtime = 1;
	}
	else
	{
	  /* Calculate number of outstanding interrupts, based on real time */

	  /* Get current time */

	  pro_time_current = pro_time_get() * pro_clk_freq;

	  pro_clk_numint = (pro_time_current - pro_time_start)/1e6;

	  /* Check number of outstanding interrupts */

	  /* Trigger int only if it is currently cleared */

	  /* XXX this is required for CATCHUP mode, but makes the emu
	     implementation slightly inaccurate */
/* XXX backwards
	     && (((pro_clk_csr2 & PRO_CLK_PF) == 0)
	        || ((pro_clk_csr2 & PRO_CLK_IRQF) == 0)))
*/

#ifndef LIMIT
	  if (pro_clk_numint < 1)
	    set_int = 0;
	  else
#endif
	  {
#ifdef LIMIT
	    if (pro_clk_numint < 1)
	    {
	      num_us = (pro_time_current - pro_time_start)/pro_clk_freq;
	      usleep(num_us);
	    }
#endif
	    set_int = 1;

	    /* Advance start time */

	    pro_time_start += 1e6;
	  }
	}

	if (set_int == 1)
#endif
	{
	  /* Trigger interrupt only if frequency is not disabled */

	  if (pro_clk_rs != 0)
	  {
	    /* Trigger interrupt */

	    pro_clk_csr2 = pro_clk_csr2 | PRO_CLK_PF;

	    if ((pro_clk_csr1 & PRO_CLK_PIE) != 0)		/* if enabled */
	    {
	      pro_clk_csr2 = pro_clk_csr2 | PRO_CLK_IRQF;
	      pro_int_set(PRO_INT_CLK);
	    }
	  }

	  /* Advance interrupt counter and update time, if required */

	  pro_clk_int_cnt++;

	  if (pro_clk_int_cnt >= pro_clk_freq)
	  {
	    pro_clk_int_cnt = 0;

	    /* Inhibit update cycles if SET bit is set */

	    if ((pro_clk_csr1 & PRO_CLK_SET) == 0)
	    {
	      /* Set update-in-progress bit */

	      pro_clk_csr0 = pro_clk_csr0 | PRO_CLK_UIP;

	      /* Schedule update-ended event */

	      pro_clk_uf_sched ();
	    }
	  }
	}

	/* Schedule next interrupt */

	pro_clk_sched();
}

/* Clock registers */

int pro_clk_rd (int pa)
{
int	data;

	switch (pa & 017777776)
	{
	  case 017773000:
	    data = pro_time_sec; /* seconds */
	    break;

	  case 017773002:
	    data = pro_clk_al_sec; /* seconds alarm */
	    break;

	  case 017773004:
	    data = pro_time_min; /* minutes */
	    break;

	  case 017773006:
	    data = pro_clk_al_min; /* minutes alarm */
	    break;

	  case 017773010:
	    data = pro_time_hour; /* hours */
	    break;

	  case 017773012:
	    data = pro_clk_al_hour; /* hours alarm */
	    break;

	  case 017773014:
	    data = pro_time->tm_wday + 1; /* day */ /* XXX correct? */
	    break;

	  case 017773016:
	    data = pro_time->tm_mday; /* date */
	    break;

	  case 017773020:
	    data = pro_time->tm_mon + 1; /* month */
	    break;

	  case 017773022:
	    /* Venix is unhappy with (70 < year > 99).
	       The following workaround allows the year to
	       be forced from the configuration file. */

	    if (pro_force_year > -1)
	      data = pro_force_year;
	    else
	      data = pro_time->tm_year % 100; /* year */
	    break;

	  case 017773024:
	    data = pro_clk_csr0;
	    break;

	  case 017773026:
	    data = pro_clk_csr1;
	    break;

	  case 017773030:
	    data = pro_clk_csr2;

	    /* clear read-once bits */

	    pro_clk_csr2 = pro_clk_csr2 & ~(PRO_CLK_IRQF | PRO_CLK_PF
	                                   | PRO_CLK_AF | PRO_CLK_UF);

	    break;

	  case 017773032:
	    data = pro_clk_csr3;

	    /* set read-once bit */

	    pro_clk_csr3 = pro_clk_csr3 | PRO_CLK_VRT;

	    break;

	  default:
	    data = 0;
	    break;
	}

	return data;
}

void pro_clk_wr (int data, int pa, int access)
{
	switch (pa & 017777776)
	{
	  case 017773002:
	    WRITE_WB(pro_clk_al_sec, PRO_CLK_AL_SEC_W, access); /* seconds alarm */
	    break;

	  case 017773006:
	    WRITE_WB(pro_clk_al_min, PRO_CLK_AL_MIN_W, access); /* minutes alarm */
	    break;

	  case 017773012:
	    WRITE_WB(pro_clk_al_hour, PRO_CLK_AL_HOUR_W, access); /* hours alarm */
	    break;

	  case 017773000:
	  case 017773004:
	  case 017773010:
	  case 017773014:
	  case 017773016:
	  case 017773020:
	  case 017773022:

	    /* Writes to time registers are not supported */
/* XXX
	    printf("%10.0lf Realtime clock write\r\n", sim_gtime());
*/
	    break;

	  case 017773024:
	    WRITE_WB(pro_clk_csr0, PRO_CLK_CSR0_W, access);
	    break;

	  case 017773026:
	    WRITE_WB(pro_clk_csr1, PRO_CLK_CSR1_W, access);
	    break;

	  case 017773030:
	  case 017773032:
	    /* read-only */
	    break;

	  default:
	    break;
	}
}


/* Reset */

void pro_clk_reset ()
{
	pro_clk_csr0 = 0000;
	pro_clk_csr1 = PRO_CLK_DM | PRO_CLK_24H;
	pro_clk_csr2 = 0000;
	pro_clk_csr3 = PRO_CLK_VRT;	/* this indicates clock has not lost power */

	pro_clk_al_sec = 0;
	pro_clk_al_min = 0;
	pro_clk_al_hour = 0;

	pro_clk_realtime = 0;
	pro_clk_numint = 0;
	pro_clk_freq = 0;
	pro_clk_int_cnt = 0;
	pro_time_start = 0;
	pro_time_current = 0;

	/* Initialize time */

	pro_time_update();

	/* Force initialization of pro_clk_freq and pro_time_start */

	pro_clk_rs = -1;

	/* Schedule first periodic event */

	pro_clk_sched();
}
#endif
